'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var S = _interopDefault(require('s-js'));

function comparer(v, k, b, isArray, path, r) {
  let index;
  const newPath = path.concat([k]);
  if (isArray && v != null && typeof v === 'object'
      && k != (index = b.findIndex(i => i && (i === v || (v.id != null && i.id === v.id) || (v._id != null && i._id === v._id)))))  {
    return r.push(newPath.concat([index > -1 ? b[index] : v]));
  }
  return r.push.apply(r, diff(v, b[k], newPath));
}

function clone(v) {
  if (!isObject(v)) return v;
  if (Array.isArray(v)) return v.slice(0);
  return Object.assign({}, v);
}

function isObject(obj) {
  let ref;
  return obj !== null && ((ref = typeof obj) === 'object' || ref === 'function');
}

function diff(a, b, path = []) {
  let i, k, l, len, v;
  const r = [];
  if (!isObject(a) || (b == null)) {
    if (a !== b) {
      r.push(path.concat([a]));
    }
  } else if (Array.isArray(a)) {
    for (k = i = 0, len = a.length; i < len; k = ++i) {
      v = a[k];
      if ((b != null ? b[k] : void 0) !== v) comparer(v, k, b, true, path, r);
    }
    if ((b != null ? b.length : void 0) > a.length) {
      l = a.length;
      while (l < b.length) {
        r.push(path.concat([l, void 0]));
        l++;
      }
    }
  } else {
    for (k in a) {
      v = a[k];
      if ((b != null ? b[k] : void 0) !== v)
        comparer(v, k, b, false, path, r);
    }
    for (k in b) {
      v = b[k];
      if (!(k in a))
        r.push(path.concat([k, void 0]));
    }
  }
  return r;
}

function unwrap(item) {
  let result, unwrapped, v;
  if (result = item != null ? item._state : void 0) return result;
  if (!isObject(item) || (typeof item === 'function') || (item instanceof Element)) return item;
  if (Object.isFrozen(item)) item = clone(item);

  if (Array.isArray(item)) {
    for (let i = 0, l = item.length; i < l; i++) {
      v = item[i];
      if ((unwrapped = unwrap(v)) !== v) item[i] = unwrapped;
    }
  } else {
    let keys = Object.keys(item);
    for (let i = 0, l = keys.length; i < l; i++) {
      v = item[keys[i]];
      if ((unwrapped = unwrap(v)) !== v) item[keys[i]] = unwrapped;
    }
  }
  return item;
}

const SNODE = Symbol('solid-node'),
  SPROXY = Symbol('solid-proxy');

const proxyTraps = {
  get(target, property) {
    if (property === '_state') return target;
    const value = target[property];
    if (S.isListening() && typeof value !== 'function') track(target, property, value);
    return (!isObject(value) || typeof value === 'function' || value instanceof Element) ? value : wrap(value);
  },

  set() { return true; },

  deleteProperty() { return true; }
};

function wrap(value) { return value[SPROXY] || (value[SPROXY] = new Proxy(value, proxyTraps)); }

function getDataNode(target) {
  let node = target[SNODE];
  if (!node) target[SNODE] = node = {};
  return node;
}

function track(target, property, value) {
  let node;
  if (isObject(value) && !(value instanceof Element)) {
    if (node = getDataNode(value)) {
      if (!node._self) node._self = S.makeDataNode();
      node._self.current();
    }
  }
  node = getDataNode(target);
  node[property] || (node[property] = S.makeDataNode());
  node[property].current();
}

function trigger(node, property, notify) {
  if (node[property]) node[property].next();
  if (notify && node._self) node._self.next();
}

function setProperty(state, property, value) {
  value = unwrap(value);
  if (state[property] === value) return;
  const notify = Array.isArray(state) || !(property in state);
  if (value === void 0) delete state[property];
  else state[property] = value;
  trigger(getDataNode(state), property, notify);
}

function mergeState(state, value) {
  const keys = Object.keys(value) || [];
  for (let i = 0; i < keys.length; i += 1) {
    const key = keys[i];
    setProperty(state, key, value[key]);
  }
}

function updatePath(current, path, traversed = []) {
  if (path.length === 1) {
    let value = path[0];
    if (typeof value === 'function') {
      value = value(wrap(current), traversed);
      // deep map
      if (Array.isArray(value)) {
        for (let i = 0; i < value.length; i += 1)
          updatePath(current, value[i], traversed);
        return;
      }
    }
    return mergeState(current, value);
  }

  const part = path.shift(),
    partType = typeof part,
    isArray = Array.isArray(current);

  if (Array.isArray(part)) {
    // Ex. update('data', [2, 23], 'label', l => l + ' !!!');
    for (let i = 0; i < part.length; i++)
      updatePath(current, [part[i]].concat(path), [].concat(traversed, [part[i]]));
  } else if (isArray && partType === 'function') {
    // Ex. update('data', i => i.id === 42, 'label', l => l + ' !!!');
    for (let i = 0; i < current.length; i++)
      if (part(current[i], i)) updatePath(current[i], path.slice(0), [].concat(traversed, [i]));
  } else if (isArray && partType === 'object') {
    // Ex. update('data', { from: 3, to: 12, by: 2 }, 'label', l => l + ' !!!');
    const {from = 0, to = current.length - 1, by = 1} = part;
    for (let i = from; i <= to; i += by)
      updatePath(current[i], path.slice(0), [].concat(traversed, [i]));
  } else if (isArray && part === '*') {
    // Ex. update('data', '*', 'label', l => l + ' !!!');
    for (let i = 0; i < current.length; i++)
      updatePath(current, [i].concat(path), [].concat(traversed, [i]));
  } else if (path.length === 1) {
    let value = path[0];
    if (typeof value === 'function')
      value = value(typeof current[part] === 'object' ? wrap(current[part]) : current[part], traversed.concat([part]));
    if (current[part] != null && typeof current[part] === 'object' && value !== null && typeof value === 'object' && !Array.isArray(value))
      return mergeState(current[part], value);
    return setProperty(current, part, value);
  } else updatePath(current[part], path);
}

function useState(state) {
  state = unwrap(state);
  const wrappedState = wrap(state);
  return [wrappedState, setState];

  function setState() {
    const args = arguments;
    S.freeze(() => {
      if (Array.isArray(args[0])) {
        for (let i = 0; i < args.length; i += 1)
          updatePath(state, args[i]);
      } else updatePath(state, Array.prototype.slice.call(args));
    });
  }
}

function reconcile() {
  const path = Array.prototype.slice.call(arguments, 0, -1),
    value = arguments[arguments.length - 1];
  return state => {
    state = unwrap(state);
    for (let i = 0; i < path.length; i += 1) state = state[path[i]];
    return diff(value, state, path);
  }
}

function useMemo(fn, seed) { return S(fn, seed); }

function useSignal(value) { return S.data(value); }

function useEffect(fn, deps, defer) {
  if (!deps) return S.effect(fn);
  S.on(deps, fn, undefined, defer);
}

function pipe(input, ...fns) {
  return compose(...fns)(input);
}

function compose(...fns) {
  if (!fns) return i => i;
  if (fns.length === 1) return fns[0];
  return input => fns.reduce(((prev, fn) => fn(prev)), input);
}

function map(fn) {
  return input => () => {
    const value = input();
    if (value === void 0) return;
    return S.sample(() => fn(value));
  }
}

function tap(fn) {
  return input => () => {
    const value = input();
    if (value !== void 0) S.sample(() => fn(value));
    return;
  }
}

// memoized map that handles falsey rejection
function when(mapFn) {
  let mapped, value, disposable;
  S.cleanup(function dispose() {
    disposable && disposable();
  });
  return map(function mapper(newValue) {
    if (newValue == null || newValue === false) {
      disposable && disposable();
      return value = mapped = disposable = null;
    }
    if (value === newValue) return mapped;
    disposable && disposable();
    disposable = null;
    value = newValue;
    return mapped = S.root((d) => {
      disposable = d;
      return mapFn(value);
    });
  })
}

// Need to be able grab wrapped state internals so can't use S-Array
function each(mapFn) {
  let mapped = [],
      list = [],
      disposables = [],
      length = 0;
  S.cleanup(function dispose() {
    for (let i = 0; i < disposables.length; i++) disposables[i]();
  });
  return map(function mapper(newList) {
    let i, j = 0,
      newLength = (newList && newList.length) || 0;
    if (newLength === 0) {
      if (length !== 0) {
        for (i = 0; i < length; i++) disposables[i]();
        list = [];
        mapped = [];
        disposables = [];
        length = 0;
      }
    } else if (length === 0) {
      j = 0;
      while (j < newLength) {
        list[j] = newList[j];
        mapped[j] = S.root(mappedFn);
        j++;
      }
      length = newLength;
    } else {
      const newMapped = new Array(newLength),
        tempDisposables = new Array(newLength),
        indexedItems = new Map();

      // reduce from both ends
      let end = Math.min(length, newLength),
        start = 0, item, itemIndex, newEnd;
      while (start < end && newList[start] === list[start]) start++;

      end = length - 1;
      newEnd = newLength - 1;
      while (end >= 0 && newEnd >= 0 && newList[newEnd] === list[end]) {
        newMapped[newEnd] = mapped[end];
        tempDisposables[newEnd] = disposables[end];
        end--;
        newEnd--;
      }

      // create indices
      j = newEnd;
      while (j >= start) {
        item = newList[j];
        itemIndex = indexedItems.get(item);
        if (itemIndex != null) itemIndex.push(j);
        else indexedItems.set(item, [j]);
        j--;
      }

      // find old items
      i = start;
      while (i <= end) {
        item = list[i];
        itemIndex = indexedItems.get(item);
        if (itemIndex != null && itemIndex.length > 0) {
          j = itemIndex.pop();
          newMapped[j] = mapped[i];
          tempDisposables[j] = disposables[i];
        } else disposables[i]();
        i++;
      }

      // set all new values
      j = start;
      while (j < newLength) {
        list[j] = newList[j];
        if (newMapped.hasOwnProperty(j)) {
          mapped[j] = newMapped[j];
          disposables[j] = tempDisposables[j];
        } else mapped[j] = S.root(mappedFn);
        j++;
      }

      // truncate extra length
      length = list.length = mapped.length = disposables.length = newLength;
    }
    return mapped;

    function mappedFn(dispose) {
      disposables[j] = dispose;
      return mapFn(list[j], j);
    }
  });
}

// export observable
function observable(input) {
  if (Symbol.observable in input) return input[Symbol.observable]();
  return {
    subscribe(observer) {
      if (!(observer instanceof Object) || observer == null) {
        throw new TypeError('Expected the observer to be an object.');
      }
      observer = observer.next || observer;
      let complete = false;
      S.on(input, function next() {
        if (complete) return;
        observer(input());
      });
      return {
        unsubscribe() { complete = true; }
      };
    },
    [Symbol.observable]() { return this; }
  };
}

const { root, cleanup: useCleanup, sample, freeze } = S;

exports.root = root;
exports.useCleanup = useCleanup;
exports.sample = sample;
exports.freeze = freeze;
exports.unwrap = unwrap;
exports.useState = useState;
exports.reconcile = reconcile;
exports.useMemo = useMemo;
exports.useSignal = useSignal;
exports.useEffect = useEffect;
exports.pipe = pipe;
exports.compose = compose;
exports.map = map;
exports.tap = tap;
exports.when = when;
exports.each = each;
exports.observable = observable;
